/****************************************************************************
 * UniPROG Universal programmer 
 * Main program
 * 
 * Daniel Quintero <danielqg@infonegocio.com>
 *
 * 06/06/2001 - First version
 * 07/09/2003 - Ported to new multi-programmer architecture
 *
 */

#include "stdhdr.h"
#include "portio.h"
#include "progdev.h"
#include "algoritms.h"
#include "intelhex.h"
#include "parseargs.h"
#include "chipdb.h"

#define BLOCK_SIZE 256

int  argLptPort = 1;
char *argChipType = "";
char *argIFilename = "";
char *argOFilename = "";
char *argProgName = "wil";
char argFormat = '\0';
int  argRead = 0;
int  argWrite = 0;
int  argVerify = 0;
int  argCheckHw = 0;
int  argErase = 0;
int  argHelp = 0;
int  argPrintList = 0;
int  argChipId = 0;
int  argVerbose = 0;
int  argReadCount = 1;
CHIPREG *chip;

ARG argTable[] = {
                  // Operations
                  { "read", BOOLEAN, "Read chip", &argRead, 0 },
                  { "write", BOOLEAN, "Write chip", &argWrite, 0 },
                  { "verify", BOOLEAN, "Verify chip", &argVerify, 0 },
                  { "erase", BOOLEAN, "Erase chip", &argErase, 0 },
                  { "getid", BOOLEAN, "Indentify chip (Flash only)", &argChipId, 0 },
		  { "checkwh", BOOLEAN, "Check Programmer", &argCheckHw, 0 },
		  { "count", INTEGER, "Read Repeat Count", &argReadCount, 0 },
                  // File options
                  { "if", STRING, "Input file", (int*) &argIFilename, 0 },
                  { "of", STRING, "Output File", (int*) &argOFilename, 0},
                  { "format", CHARACTER,  "Data Format [h]Intel Hex [b]Binary", (int*) &argFormat, 0 },
                  // Device
                  { "chip", STRING,  "Chip type (use -printdb for complete list)", (int*) &argChipType, 0 },
                  // Interface
                  { "lpt", INTEGER, "LPT port number ( 1 is LPT1 )", &argLptPort, 0 },
		  { "prog", STRING, "Programmer hardware", (int*) &argProgName, 0 },
                  // Information
                  { "printdb", BOOLEAN, "Print list of supported chips", &argPrintList, 0 },
                  { "verbose", INTEGER, "Verbosity level (0, 1)", &argVerbose, 0},
		  { "?,h,help", BOOLEAN, "Print help", &argHelp },
                  };

#define NUM_ARGS (sizeof(argTable) / sizeof(ARG))



typedef char *pchar;
const pchar errTable[] = {
                  "No error",           // AL_NOERR
                  "Timeout exceded",    // AL_ERRTIMEOUT
                  "Verify failed",      // AL_ERRVERIFY
                  "Unknown error"
                  };

#define NUM_ERR (sizeof(errTable) / sizeof(char *))



void printError( int err )
  {
  if (err < NUM_ERR-1)
    printf("Error : %s\n", errTable[err]);
  else
    printf("Error : %s\n", errTable[err+1]);
  }


void printHex(unsigned char *buf, int sizebuf)
  {
  int i;
  unsigned char *p = buf;

  for (i=0; i < sizebuf; i++)
    printf("%02X ", *p++);

  for (i=0, p = buf; i < sizebuf; i++, p++)
    putchar( isprint(*p) ? *p : '.');
  putchar('\n');
  }

/* UNIX console setup functions */
#ifdef __LINUX__

static struct termios stored_settings;

void canon_off(void)
{
    struct termios new_settings;
    tcgetattr(0,&stored_settings);
    new_settings = stored_settings;
    new_settings.c_lflag &= (~ICANON) & (~ECHO);
    new_settings.c_cc[VTIME] = 0;
    new_settings.c_cc[VMIN] = 1;
    tcsetattr(0,TCSANOW,&new_settings);
    return;
}

void canon_on(void)
{
    tcsetattr(0,TCSANOW,&stored_settings);
    return;
}

#endif /*__LINUX__*/




void ReadChip( unsigned char *buf, int size_chip )
  {
  int pos = 0;
  unsigned long addr;
  unsigned char *p;

  printf("Reading chip ");

  /* Power up */
  genVccOn();
  for (addr=0, p=buf; addr < size_chip; addr+=BLOCK_SIZE, p+=BLOCK_SIZE)
    {
    genReadBlock(p, BLOCK_SIZE, addr);
    if (addr > (size_chip / 60)*pos)
      {
      printf("#");
      pos++;
      }
    }
  printf(" OK \n");

  /* Power down */
  genVccOff();
  }


void IdentifyChip( void )
  {
  unsigned int id;
  CHIPREG *chip;

  genVccOn();
  atGetID(&id);
  genVccOff();

  printf("Chip ID : %04X\n", id);
  chip = chipdbFindId( id );
  printf("Vendor  : %s\n", (chip) ? chip->brand : 
	 chipdbGetVendorName((id >> 8) & 0xff) );
  printf("Device  : %s\n", (chip) ? chip->name  : "Unknow" );

  if ((id >> 8) == 0x01) { /* AMD */
    genVccOn();
    amGetSectorProtection();
    genVccOff();
  }
  }


void WriteChip(unsigned char *buf, int size, CHIPREG *chip)
  {
  int pos = 0;
  unsigned long addr;
  int err;
  unsigned int sector_size = (chip->size * 1024) / chip->nSectors;
  unsigned char *p;

  printf("Writing chip ");

  /* Power up chip */
  genVccOn();
  for (addr=0, p=buf; addr < size; addr+=sector_size, p+=sector_size)
    {
    int retry = 2;                      // max number of retries before error
    unsigned int n = size - addr;       // bytes for this sector
    n = (n < sector_size) ? n : sector_size;

    write_sector:
    /* write the sector */
    err = atWriteSector(p, n, addr, chip);
    if (err != AL_NOERR) {
      //printError(err);
      //      goto error;
      }

    /* verify the sector */
    err = atVerifySector(p, n, addr, chip);
    if (err != AL_NOERR) {
      if (retry--) {
        printf("R");
        goto write_sector;
        }
      printError(err);
      goto error;
      }

    /* print a nice progress bar */
    if (addr > (size / 60)*pos)
      {
      printf("#");
      pos++;
      }
    }
  printf(" OK \n");
  err = AL_NOERR;

  error : 
    
  /* Power down chip */
  genVccOff();
  }


void VerifyChip(unsigned char *buf, int size, CHIPREG *chip)
  {
  int pos = 0;
  unsigned long addr;
  int err;
  unsigned int sector_size;
  unsigned char *p;

  if (chip->nSectors)
    sector_size = (chip->size * 1024) / chip->nSectors;
  else
    sector_size = BLOCK_SIZE;

  printf("Verifing chip ");

  /* Power up chip */
  genVccOn();

  for (addr=0, p=buf; addr < size; addr+=sector_size, p+=sector_size)
    {
    unsigned int n = size - addr;       // bytes for this sector
    n = (n < sector_size) ? n : sector_size;

    /* verify the sector */
    err = atVerifySector(p, n, addr, chip);
    if (err != AL_NOERR) {
      printError(err);
      goto error;
      }

    /* print a nice progress bar */
    if (addr > (size / 60)*pos)
      {
      printf("#");
      pos++;
      }
    }
  printf(" OK \n");

  error:
  
  /* Power off chip */
  genVccOff();
  }


/******************************************************************
  Operations
 ******************************************************************/

// Read Chip

int opReadChip( void ) {
  // allocate memory for buffer
  unsigned long chipsize = chip->size*1024L;
  char *buf = malloc(chipsize);
  int err;

  ReadChip(buf, chipsize);

  // Save Data
  printf("Saving %ldKb of data to \"%s\"\n", chip->size, argOFilename);
  err = fileSave(argOFilename, buf, chipsize, argFormat);
  if (err < 0) {
    fprintf(stderr, "Error : Can't save file\n");
    return -1;
    }

  // free memory
  free(buf);
  return 0;
  }

// Write Chip

int opWriteChip( void ) {
  // allocate memory for buffer
  unsigned long fsize, chipsize = chip->size*1024L;
  char *buf = malloc(chipsize);
  int err;
  fsize = chipsize;

  // default value for data is 0xff
  memset(buf, 0xFF, chipsize);

  // Load Data
  printf("Loading data from \"%s\"\n", argIFilename);
  err = fileLoad(argIFilename, buf, chipsize, &fsize, argFormat);
  if (err < 0) {
    fprintf(stderr, "Error : Can't load file\n");
    return -1;
    }

  // write chip
  WriteChip( buf, fsize, chip);

  // free memory
  free(buf);
  return 0;
  }

int opVerifyChip( void ) {
  // allocate memory for buffer
  unsigned long fsize, chipsize = chip->size*1024L;
  char *buf = malloc(chipsize);
  int err;
  fsize = chipsize;

  // default value for data is 0
  memset(buf, 0, chipsize);

  // Load Data
  printf("Loading data from \"%s\"\n", argIFilename);
  err = fileLoad(argIFilename, buf, chipsize, &fsize, argFormat);
  if (err < 0) {
    fprintf(stderr, "Error : Can't load file\n");
    return -1;
    }

  // verify chip
  VerifyChip( buf, fsize, chip);

  // free memory
  free(buf);
  return 0;
  }

int opCheckHw( void ) {
  CheckHardware();
  return 0;
  }

int opEraseChip( void ) {
  /* Power up chip */
  genVccOn();

  /* Chip erase */
  atErase();

  /* Power off chip */
  genVccOn();
  return 0;
  }


int main(int argc, char* argv[]) {
  int err = 0;
  int hwOpen = FALSE;
  struct prog_dev *pd = NULL;
  
#ifdef __WIN32__
  /* This is a must for WINNT */
  SetPriorityClass(GetCurrentThread(), REALTIME_PRIORITY_CLASS);
  SetThreadPriority(GetCurrentThread(), THREAD_PRIORITY_TIME_CRITICAL);
#endif
  
#ifdef __LINUX__
  /* On linux console, disable canonical mode */
  canon_off();
#endif

  /* Print Banner */
  printf("UniProg v0.1b, Daniel Quintero <danielqg@infonegocio.com>, 8-Sep-03\n");
 
  /* load programmers modules */
  prgdev_init();

  /* parse arguments */
  if (argParse( argc, argv, argTable, NUM_ARGS)) {
    err = 1;
    goto exit;
  }

  if (argHelp || argc < 2) {
    fprintf(stderr, "Legal arguments are:\n\n");
    argPrintUsage(argTable, NUM_ARGS);
    err = 1;
    goto exit;
  }
  
  if (argPrintList) {
    printf("Supported chips :\n\n");
    chipdbPrintList();
    err = 0;
    goto exit;
  }
  
  
  if (argVerbose > 0)  {
    printf("\n\nParallel port : LPT%d\n", argLptPort);
    printf("In File       : %s\n", argIFilename);
    printf("Out File      : %s\n", argOFilename);
    printf("Format        : %c\n", argFormat);
    printf("Read chip     : %s\n", argRead ? "yes" : "no");
    printf("Write chip    : %s\n", argWrite ? "yes" : "no");
    printf("Verify chip   : %s\n", argVerify ? "yes" : "no");
    printf("Erase chip    : %s\n\n", argErase ? "yes" : "no");
  }


  /* Init Programmer, TODO: close programmer at exit */
  prgdev_select_device(argProgName);
  pd = prgdev_get_selected();
  assert(pd);
  printf("Using hardware %s\n", pd->description);

  if (pd->open(pd) != ERR_OK) {
    fprintf(stderr, "Error, programmer not found\n"); 
    err = 1;
    goto exit;
  }
  else
    hwOpen = TRUE;
  
  /* Programmer operations */

  /* Hardware testbench */
  if (argCheckHw)  {  
    opCheckHw();
    goto exit;
  }

  /* Identify chip (only for flash) */
  if (argChipId) {
    IdentifyChip();
    goto exit;
  }
 
  /* Chip access operations */

  if (!(argRead || argErase || argWrite || argVerify))  {
    fprintf(stderr, "no operation requested\n"); 
    err = 1;
    goto exit;
  }

  /* Check for chip argument */
  if (argChipType == NULL) {
    fprintf(stderr, "Chip parameter is required for read, write or verify operations\n");
    err = 1;
    goto exit;
  }
  
  if (((chip = chipdbFind(argChipType)) == NULL)) {
    fprintf(stderr, "Unknown chip %s, use printdb parameter to see supported chips\n",
	    argChipType);
    err = 1;
    goto exit;
  }
  
  printf("Chip Type     : %s - %s\n", chip->name, chip->desc);

  /* Read operation */
  if (argRead)   {
    if (argWrite | argErase | argVerify) {
      fprintf(stderr, "read is incompatible with write, erase or verify actions\n");
      goto exit;
    }
    
  if (*argOFilename == '\0') {
    fprintf(stderr, "output file don't specified\n");
    goto exit;
  }

  /* if format not specified, try to autodetect */
  if (argFormat == '\0')
    if ((argFormat = getFilenameFormat(argOFilename)) == '\0') {
      fprintf(stderr, "Unknown file format, please use format parameter\n"); 
      err = 1;
      goto exit;
    }
  
  err = opReadChip();
  }

  /* check parameters to erase/write/verify */
  if (argWrite) {
    if (*argIFilename == '\0') {
      fprintf(stderr, "input file don't specified\n");
      goto exit;
    }
  
  /* if format not specified, try to autodetect */
    if (argFormat == '\0')
      if ((argFormat = getFilenameFormat(argIFilename)) == '\0') {
	fprintf(stderr, "Unknown file format, please use format parameter\n"); 
	err = 1;
	goto exit;
      }
  } 

  /* Erase */
  if (argErase) 
    err = opEraseChip();
  if (err) goto exit;

  /* Write */
  if (argWrite)
    err = opWriteChip();
  if (err) goto exit;

  /* Verify */
  if (argVerify)
    err = opVerifyChip();
  if (err) goto exit;

 exit:
  
#ifdef __LINUX__
  /* restore UNIX console settings */
  canon_on();
#endif
  
  if (hwOpen)
    pd->close(pd);
  return err;
}
//---------------------------------------------------------------------------
